from __future__ import absolute_import

import wrapt

from ..log import logger
from ..singletons import tracer

try:
    import redis

    def collect_tags(span, instance, args, kwargs):
        try:
            ckw = instance.connection_pool.connection_kwargs

            span.set_tag("driver", "redis-py")

            host = ckw.get('host', None)
            port = ckw.get('port', '6379')
            db = ckw.get('db', None)

            if host is not None:
                url = "redis://%s:%s" % (host, port)
                if db is not None:
                    url = url + "/%s" % db
                span.set_tag('connection', url)

        except:
            logger.debug("redis.collect_tags non-fatal error", exc_info=True)
        finally:
            return span


    def execute_command_with_instana(wrapped, instance, args, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None or parent_span.operation_name == "redis":
            return wrapped(*args, **kwargs)

        with tracer.start_active_span("redis", child_of=parent_span) as scope:
            try:
                collect_tags(scope.span, instance, args, kwargs)
                if (len(args) > 0):
                    scope.span.set_tag("command", args[0])

                rv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv


    def execute_with_instana(wrapped, instance, args, kwargs):
        parent_span = tracer.active_span

        # If we're not tracing, just return
        if parent_span is None or parent_span.operation_name == "redis":
            return wrapped(*args, **kwargs)

        with tracer.start_active_span("redis", child_of=parent_span) as scope:
            try:
                collect_tags(scope.span, instance, args, kwargs)
                scope.span.set_tag("command", 'PIPELINE')

                pipe_cmds = []
                for e in instance.command_stack:
                    pipe_cmds.append(e[0][0])
                scope.span.set_tag("subCommands", pipe_cmds)
            except Exception as e:
                # If anything breaks during K/V collection, just log a debug message
                logger.debug("Error collecting pipeline commands", exc_info=True)

            try:
                rv = wrapped(*args, **kwargs)
            except Exception as e:
                scope.span.log_exception(e)
                raise
            else:
                return rv

    if redis.VERSION < (3,0,0):
        wrapt.wrap_function_wrapper('redis.client', 'BasePipeline.execute', execute_with_instana)
        wrapt.wrap_function_wrapper('redis.client', 'StrictRedis.execute_command', execute_command_with_instana)
    else:
        wrapt.wrap_function_wrapper('redis.client', 'Pipeline.execute', execute_with_instana)
        wrapt.wrap_function_wrapper('redis.client', 'Redis.execute_command', execute_command_with_instana)

        logger.debug("Instrumenting redis")
except ImportError:
    pass



from basictracer.span import BasicSpan
from .log import logger


class InstanaSpan(BasicSpan):
    stack = None

    def finish(self, finish_time=None):
        super(InstanaSpan, self).finish(finish_time)

    def log_exception(self, e):
        try:
            message = ""

            self.set_tag("error", True)
            ec = self.tags.get('ec', 0)
            self.set_tag("ec", ec+1)

            if hasattr(e, '__str__'):
                message = str(e)
            elif hasattr(e, 'message') and e.message is not None:
                message = e.message

            if self.operation_name in ['rpc-server', 'rpc-client']:
                self.set_tag('rpc.error', message)

            self.log_kv({'message': message})

        except Exception:
            logger.debug("span.log_exception", exc_info=True)
            raise




from __future__ import absolute_import

import wrapt

from ..log import logger
from ..singletons import async_tracer
from ..configurator import config

try:
    import asyncio

    @wrapt.patch_function_wrapper('asyncio','ensure_future')
    def ensure_future_with_instana(wrapped, instance, argv, kwargs):
        if config['asyncio_task_context_propagation']['enabled'] is False:
            return wrapped(*argv, **kwargs)

        scope = async_tracer.scope_manager.active
        task = wrapped(*argv, **kwargs)

        if scope is not None:
            async_tracer.scope_manager._set_task_scope(scope, task=task)

        return task

    if hasattr(asyncio, "create_task"):
        @wrapt.patch_function_wrapper('asyncio','create_task')
        def create_task_with_instana(wrapped, instance, argv, kwargs):
            if config['asyncio_task_context_propagation']['enabled'] is False:
                return wrapped(*argv, **kwargs)

            scope = async_tracer.scope_manager.active
            task = wrapped(*argv, **kwargs)

            if scope is not None:
                async_tracer.scope_manager._set_task_scope(scope, task=task)

            return task

    logger.debug("Instrumenting asyncio")
except ImportError:
    pass



"""
The uwsgi and uwsgidecorators packages are added automatically to the Python environment
when running under uWSGI.  Here we attempt to detect the presence of these packages and
then use the appropriate hooks.
"""
from __future__ import absolute_import

from ..log import logger
from ..singletons import agent

try:
    import uwsgi
    logger.debug("uWSGI options: %s", uwsgi.opt)

    opt_master = uwsgi.opt.get('master', False)
    opt_lazy_apps = uwsgi.opt.get('lazy-apps', False)

    if uwsgi.opt.get('enable-threads', False) is False:
        logger.warn("Required: uWSGI threads are not enabled.  " +
                    "Please enable by using the uWSGI --enable-threads option.")

    if opt_master and opt_lazy_apps is False:
        # --master is supplied in uWSGI options (otherwise uwsgidecorators package won't be available)
        # When --lazy-apps is True, this postfork hook isn't needed
        import uwsgidecorators

        @uwsgidecorators.postfork
        def uwsgi_handle_fork():
            """ This is our uWSGI hook to detect and act when worker processes are forked off. """
            logger.debug("Handling uWSGI fork...")
            agent.handle_fork()

        logger.debug("Applied uWSGI hooks")
    else:
        logger.debug("uWSGI --master=%s --lazy-apps=%s: postfork hooks not applied", opt_master, opt_lazy_apps)
except ImportError as e:
    logger.debug('uwsgi hooks: decorators not available: %s', e)
    pass



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import tests.apps.grpc_server.stan_pb2 as stan__pb2


class StanStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.OneQuestionOneResponse = channel.unary_unary(
        '/stan.Stan/OneQuestionOneResponse',
        request_serializer=stan__pb2.QuestionRequest.SerializeToString,
        response_deserializer=stan__pb2.QuestionResponse.FromString,
        )
    self.ManyQuestionsOneResponse = channel.stream_unary(
        '/stan.Stan/ManyQuestionsOneResponse',
        request_serializer=stan__pb2.QuestionRequest.SerializeToString,
        response_deserializer=stan__pb2.QuestionResponse.FromString,
        )
    self.OneQuestionManyResponses = channel.unary_stream(
        '/stan.Stan/OneQuestionManyResponses',
        request_serializer=stan__pb2.QuestionRequest.SerializeToString,
        response_deserializer=stan__pb2.QuestionResponse.FromString,
        )
    self.ManyQuestionsManyReponses = channel.stream_stream(
        '/stan.Stan/ManyQuestionsManyReponses',
        request_serializer=stan__pb2.QuestionRequest.SerializeToString,
        response_deserializer=stan__pb2.QuestionResponse.FromString,
        )
    self.OneQuestionOneErrorResponse = channel.unary_unary(
        '/stan.Stan/OneQuestionOneErrorResponse',
        request_serializer=stan__pb2.QuestionRequest.SerializeToString,
        response_deserializer=stan__pb2.QuestionResponse.FromString,
        )
    self.OneErroredQuestionOneResponse = channel.unary_unary(
        '/stan.Stan/OneErroredQuestionOneResponse',
        request_serializer=stan__pb2.QuestionRequest.SerializeToString,
        response_deserializer=stan__pb2.QuestionResponse.FromString,
        )


class StanServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def OneQuestionOneResponse(self, request, context):
    """Unary
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ManyQuestionsOneResponse(self, request_iterator, context):
    """Streaming
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def OneQuestionManyResponses(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ManyQuestionsManyReponses(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def OneQuestionOneErrorResponse(self, request, context):
    """Error Testing
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def OneErroredQuestionOneResponse(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_StanServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'OneQuestionOneResponse': grpc.unary_unary_rpc_method_handler(
          servicer.OneQuestionOneResponse,
          request_deserializer=stan__pb2.QuestionRequest.FromString,
          response_serializer=stan__pb2.QuestionResponse.SerializeToString,
      ),
      'ManyQuestionsOneResponse': grpc.stream_unary_rpc_method_handler(
          servicer.ManyQuestionsOneResponse,
          request_deserializer=stan__pb2.QuestionRequest.FromString,
          response_serializer=stan__pb2.QuestionResponse.SerializeToString,
      ),
      'OneQuestionManyResponses': grpc.unary_stream_rpc_method_handler(
          servicer.OneQuestionManyResponses,
          request_deserializer=stan__pb2.QuestionRequest.FromString,
          response_serializer=stan__pb2.QuestionResponse.SerializeToString,
      ),
      'ManyQuestionsManyReponses': grpc.stream_stream_rpc_method_handler(
          servicer.ManyQuestionsManyReponses,
          request_deserializer=stan__pb2.QuestionRequest.FromString,
          response_serializer=stan__pb2.QuestionResponse.SerializeToString,
      ),
      'OneQuestionOneErrorResponse': grpc.unary_unary_rpc_method_handler(
          servicer.OneQuestionOneErrorResponse,
          request_deserializer=stan__pb2.QuestionRequest.FromString,
          response_serializer=stan__pb2.QuestionResponse.SerializeToString,
      ),
      'OneErroredQuestionOneResponse': grpc.unary_unary_rpc_method_handler(
          servicer.OneErroredQuestionOneResponse,
          request_deserializer=stan__pb2.QuestionRequest.FromString,
          response_serializer=stan__pb2.QuestionResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'stan.Stan', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



from __future__ import absolute_import

import wrapt
import logging
import sys

from ..log import logger
from ..singletons import tracer


@wrapt.patch_function_wrapper('logging', 'Logger._log')
def log_with_instana(wrapped, instance, argv, kwargs):
    # argv[0] = level
    # argv[1] = message
    # argv[2] = args for message
    try:
        parent_span = tracer.active_span

        # Only needed if we're tracing and serious log
        if parent_span and argv[0] >= logging.WARN:
            # get the formatted log message
            # clients such as suds-jurko log things such as: Fault(Server: 'Server side fault example.')
            # So make sure we're working with a string
            msg = str(argv[1]) % argv[2]

            # get additional information if an exception is being handled
            parameters = None
            (t, v, tb) = sys.exc_info()
            if t is not None and v is not None:
                parameters = '{} {}'.format(t , v)

            # create logging span
            with tracer.start_active_span('log', child_of=parent_span) as scope:
                scope.span.log_kv({ 'message': msg })
                if parameters is not None:
                    scope.span.log_kv({ 'parameters': parameters })
                # extra tags for an error
                if argv[0] >= logging.ERROR:
                    scope.span.set_tag('error', True)
                    ec = scope.span.tags.get('ec', 0)
                    scope.span.set_tag('ec', ec + 1)
    except Exception as e:
        logger.debug('Exception: %s', e, exc_info=True)
    finally:
        return wrapped(*argv, **kwargs)


logger.debug('Instrumenting logging')




"""
cassandra instrumentation
https://docs.datastax.com/en/developer/python-driver/3.20/
https://github.com/datastax/python-driver
"""
from __future__ import absolute_import

from distutils.version import LooseVersion
import wrapt

from ..log import logger
from ..singletons import tracer

try:
    import cassandra

    consistency_levels = dict({0: "ANY",
                               1: "ONE",
                               2: "TWO",
                               3: "THREE",
                               4: "QUORUM",
                               5: "ALL",
                               6: "LOCAL_QUORUM",
                               7: "EACH_QUORUM",
                               8: "SERIAL",
                               9: "LOCAL_SERIAL",
                               10: "LOCAL_ONE"})

    def collect_response(span, fn):
        tried_hosts = list()
        for host in fn.attempted_hosts:
            tried_hosts.append("%s:%d" % (host.endpoint.address, host.endpoint.port))

        span.set_tag("cassandra.triedHosts", tried_hosts)
        span.set_tag("cassandra.coordHost", fn.coordinator_host)

        cl = fn.query.consistency_level
        if cl and cl in consistency_levels:
            span.set_tag("cassandra.achievedConsistency", consistency_levels[cl])


    def cb_request_finish(results, span, fn):
        collect_response(span, fn)
        span.finish()

    def cb_request_error(results, span, fn):
        collect_response(span, fn)

        span.set_tag("error", True)
        ec = span.tags.get('ec', 0)
        span.set_tag("ec", ec + 1)
        span.set_tag("cassandra.error", results.message)
        span.finish()

    def request_init_with_instana(fn):
        parent_span = tracer.active_span

        if parent_span is not None:
            ctags = dict()
            if isinstance(fn.query, cassandra.query.SimpleStatement):
                ctags["cassandra.query"] = fn.query.query_string
            elif isinstance(fn.query, cassandra.query.BoundStatement):
                ctags["cassandra.query"] = fn.query.prepared_statement.query_string

            ctags["cassandra.keyspace"] = fn.session.keyspace
            ctags["cassandra.cluster"] = fn.session.cluster.metadata.cluster_name

            span = tracer.start_span(
                operation_name="cassandra",
                child_of=parent_span,
                tags=ctags)

            fn.add_callback(cb_request_finish, span, fn)
            fn.add_errback(cb_request_error, span, fn)

    @wrapt.patch_function_wrapper('cassandra.cluster', 'Session.__init__')
    def init_with_instana(wrapped, instance, args, kwargs):
        session =  wrapped(*args, **kwargs)
        instance.add_request_init_listener(request_init_with_instana)
        return session

    logger.debug("Instrumenting cassandra")

except ImportError:
    pass


"""
The Instana package has two core components: the agent and the tracer.

The agent is individual to each python process and handles process metric
collection and reporting.

The tracer upholds the OpenTracing API and is responsible for reporting
span data to Instana.

The following outlines the hierarchy of classes for these two components.

Agent
  Sensor
    Meter

Tracer
  Recorder
"""

from __future__ import absolute_import

import os
import sys
from threading import Timer
import pkg_resources

__author__ = 'Instana Inc.'
__copyright__ = 'Copyright 2019 Instana Inc.'
__credits__ = ['Pavlo Baron', 'Peter Giacomo Lombardo']
__license__ = 'MIT'
__maintainer__ = 'Peter Giacomo Lombardo'
__email__ = 'peter.lombardo@instana.com'

try:
    __version__ = pkg_resources.get_distribution('instana').version
except pkg_resources.DistributionNotFound:
    __version__ = 'unknown'


def load(_):
    """
    Method used to activate the Instana sensor via AUTOWRAPT_BOOTSTRAP
    environment variable.
    """
    if "INSTANA_DEBUG" in os.environ:
        print("Instana: activated via AUTOWRAPT_BOOTSTRAP")


def boot_agent():
    """Initialize the Instana agent and conditionally load auto-instrumentation."""
    # Disable all the unused-import violations in this function
    # pylint: disable=unused-import

    import instana.singletons

    # Instrumentation
    if "INSTANA_DISABLE_AUTO_INSTR" not in os.environ:
        # Import & initialize instrumentation
        if sys.version_info >= (3, 5, 3):
            from .instrumentation import asyncio
            from .instrumentation.aiohttp import client
            from .instrumentation.aiohttp import server
            from .instrumentation import asynqp

        if sys.version_info[0] < 3:
            from .instrumentation import mysqlpython
            from .instrumentation import webapp2_inst
        else:
            from .instrumentation import mysqlclient

        from .instrumentation import cassandra_inst
        from .instrumentation import couchbase_inst
        from .instrumentation import flask
        from .instrumentation import grpcio
        from .instrumentation.tornado import client
        from .instrumentation.tornado import server
        from .instrumentation import logging
        from .instrumentation import pymysql
        from .instrumentation import psycopg2
        from .instrumentation import redis
        from .instrumentation import sqlalchemy
        from .instrumentation import sudsjurko
        from .instrumentation import urllib3
        from .instrumentation.django import middleware

    # Hooks
    from .hooks import hook_uwsgi


if "INSTANA_MAGIC" in os.environ:
    pkg_resources.working_set.add_entry("/tmp/instana/python")

    if "INSTANA_DEBUG" in os.environ:
        print("Instana: activated via AutoTrace")
else:
    if ("INSTANA_DEBUG" in os.environ) and ("AUTOWRAPT_BOOTSTRAP" not in os.environ):
        print("Instana: activated via manual import")

# User configurable EUM API key for instana.helpers.eum_snippet()
# pylint: disable=invalid-name
eum_api_key = ''

# This Python package can be loaded into Python processes one of three ways:
#   1. manual import statement
#   2. autowrapt hook
#   3. dynamically injected remotely
#
# With such magic, we may get pulled into Python processes that we have no interest being in.
# As a safety measure, we maintain a "do not load list" and if this process matches something
# in that list, then we go sit in a corner quietly and don't load anything at all.
do_not_load_list = ["pip", "pip2", "pip3", "pipenv", "docker-compose", "easy_install", "easy_install-2.7",
                    "smtpd.py", "twine", "ufw", "unattended-upgrade"]

# There are cases when sys.argv may not be defined at load time.  Seems to happen in embedded Python,
# and some Pipenv installs.  If this is the case, it's best effort.
if hasattr(sys, 'argv') and len(sys.argv) > 0 and (os.path.basename(sys.argv[0]) in do_not_load_list):
    if "INSTANA_DEBUG" in os.environ:
        print("Instana: No use in monitoring this process type (%s).  Will go sit in a corner quietly." % os.path.basename(sys.argv[0]))
else:
    if "INSTANA_MAGIC" in os.environ:
        # If we're being loaded into an already running process, then delay agent initialization
        t = Timer(2.0, boot_agent)
        t.start()
    else:
        boot_agent()
